"""Errors thrown by icalendar."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator


class InvalidCalendar(ValueError):
    """The calendar given is not valid.

    This calendar does not conform with RFC 5545 or breaks other RFCs.
    """


class ICalParsingError(InvalidCalendar):
    """Could not parse an iCalendar."""

    def __init__(
        self,
        message: str,
        line: str | None = None,
        line_number: int | None = None,
        value: object = None,
    ) -> None:
        self.message = message
        self.line = line
        self.line_number = line_number
        self.value = value

        full_message = message

        if value is not None:
            full_message += f": {value!r}"

        if line_number is not None and line is not None:
            full_message += f" (line {line_number}: {line!r})"
        elif line_number is not None:
            full_message += f" (line {line_number})"
        elif line is not None:
            full_message += f" ({line!r})"

        super().__init__(full_message)


class BrokenCalendarProperty(InvalidCalendar):
    """A property could not be parsed and its value is broken.

    This error is raised when accessing attributes on a
    :class:`~icalendar.prop.vBroken` property that would normally
    be present on the expected type. The original parse error
    is chained as ``__cause__``.
    """


class IncompleteComponent(ValueError):
    """The component is missing attributes.

    The attributes are not required, otherwise this would be
    an InvalidCalendar. But in order to perform calculations,
    this attribute is required.

    This error is not raised in the UPPERCASE properties like .DTSTART,
    only in the lowercase computations like .start.
    """


class IncompleteAlarmInformation(ValueError):
    """The alarms cannot be calculated yet because information is missing."""


class LocalTimezoneMissing(IncompleteAlarmInformation):
    """We are missing the local timezone to compute the value.

    Use Alarms.set_local_timezone().
    """


class ComponentEndMissing(IncompleteAlarmInformation):
    """We are missing the end of a component that the alarm is for.

    Use Alarms.set_end().
    """


class ComponentStartMissing(IncompleteAlarmInformation):
    """We are missing the start of a component that the alarm is for.

    Use Alarms.set_start().
    """


class FeatureWillBeRemovedInFutureVersion(DeprecationWarning):
    """This feature will be removed in a future version."""


class GloballyUniqueTZIDGuessed(UserWarning):
    """A globally unique TZID was resolved by stripping the vendor prefix.

    Per :rfc:`5545#section-3.2.19`, the trailing component is convention only
    and not guaranteed to match a known Olson identifier. Suppress this warning
    if the resolved timezone is correct for your data.
    """


def _repr_index(index: str | int) -> str:
    """Create a JSON compatible representation for the index.

    Parameters:
        index: It is either a dict key (string) or a list position (integer).

    Returns:
        The index as a quoted string if it's a string, else the string
        representation if it's an integer.
    """
    if isinstance(index, str):
        return f'"{index}"'
    return str(index)


class JCalParsingError(InvalidCalendar):
    """Could not parse a part of the JCal."""

    _default_value = object()

    def __init__(
        self,
        message: str,
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
        value: object = _default_value,
    ) -> None:
        """Create a new JCalParsingError.

        Parameters:
            message: A description of the error that occurred while parsing.
            parser: The parser class or its name where the error occurred.
            path: The location in the jCal structure where the error occurred.
            value: The value which caused the error, if available.
        """
        self.path = self._get_path(path)
        if not isinstance(parser, str):
            parser = parser.__name__
        self.parser = parser
        self.message = message
        self.value = value
        full_message = message
        repr_path = ""
        if self.path:
            repr_path = "".join([f"[{_repr_index(index)}]" for index in self.path])
            full_message = f"{repr_path}: {full_message}"
            repr_path += " "
        if parser:
            full_message = f"{repr_path}in {parser}: {message}"
        if value is not self._default_value:
            full_message += f" Got value: {value!r}"
        super().__init__(full_message)

    @classmethod
    @contextlib.contextmanager
    def reraise_with_path_added(
        cls,
        *path_components: int | str,
    ) -> Generator[None, None, None]:
        """Automatically re-raise the exception with path components added.

        Raises:
            ~error.JCalParsingError: If there was an exception in the context.
        """
        try:
            yield
        except JCalParsingError as e:
            raise cls(
                path=list(path_components) + e.path,
                parser=e.parser,
                message=e.message,
                value=e.value,
            ).with_traceback(e.__traceback__) from e

    @staticmethod
    def _get_path(path: list[str | int] | None | str | int) -> list[str | int]:
        """Return the path as a list."""
        if path is None:
            path = []
        elif not isinstance(path, list):
            path = [path]
        return path

    @classmethod
    def validate_property(
        cls,
        jcal_property: list[object],
        parser: str | type,
        path: list[str | int] | None | str | int = None,
    ) -> None:
        """Validate a jCal property.

        Parameters:
            jcal_property: A list with at least four items (name,
                parameters, value type, and value) which is the jCal property
                to be validated.
            parser: The parser class or its name where the error occurred.
            path: The location in the jCal structure where the error occurred.

        Raises:
            ~error.JCalParsingError: if the property is not valid.
        """
        path = cls._get_path(path)
        if not isinstance(jcal_property, list) or len(jcal_property) < 4:
            raise JCalParsingError(
                "The property must be a list with at least 4 items.",
                parser,
                path,
                value=jcal_property,
            )
        if not isinstance(jcal_property[0], str):
            raise JCalParsingError(
                "The name must be a string.", parser, path + [0], value=jcal_property[0]
            )
        if not isinstance(jcal_property[1], dict):
            raise JCalParsingError(
                "The parameters must be a mapping.",
                parser,
                path + [1],
                value=jcal_property[1],
            )
        if not isinstance(jcal_property[2], str):
            raise JCalParsingError(
                "The VALUE parameter must be a string.",
                parser,
                path + [2],
                value=jcal_property[2],
            )
        cls.validate_jcal_token(jcal_property[0], "property name", parser, path + [0])

    _type_names = {
        str: "a string",
        int: "an integer",
        float: "a float",
        bool: "a boolean",
    }

    @classmethod
    def validate_value_type(
        cls,
        jcal: object,
        expected_type: type[str | int | float | bool]
        | tuple[type[str | int | float | bool], ...],
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
    ) -> None:
        """Validate the type of a jCal value."""
        if not isinstance(jcal, expected_type):
            type_name = (
                cls._type_names[expected_type]
                if isinstance(expected_type, type)
                else " or ".join(cls._type_names[t] for t in expected_type)
            )
            raise cls(
                f"The value must be {type_name}.",
                parser=parser,
                value=jcal,
                path=path,
            )

    @classmethod
    def validate_list_type(
        cls,
        jcal: object,
        expected_type: type[str | int | float | bool],
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
    ) -> None:
        """Validate the type of each item in a jCal list."""
        path = cls._get_path(path)
        if not isinstance(jcal, list):
            raise cls(
                "The value must be a list.",
                parser=parser,
                value=jcal,
                path=path,
            )
        for index, item in enumerate(jcal):
            if not isinstance(item, expected_type):
                type_name = cls._type_names[expected_type]
                raise cls(
                    f"Each item in the list must be {type_name}.",
                    parser=parser,
                    value=item,
                    path=path + [index],
                )

    @classmethod
    def validate_jcal_token(
        cls,
        name: str,
        kind: str,
        parser: str | type = "",
        path: list[str | int] | None | str | int = None,
    ) -> None:
        r"""Validate a jCal ``name`` as a lowercase iCalendar token.

        jCal keeps a property name, parameter name, or ``RRULE`` part name
        verbatim and re-emits it into the content line on serialization, so a
        ``:``, ``;``, or lone carriage return in the name could inject
        parameters or a new content line. A valid name matches the iCalendar
        token pattern ``[\w.-]+`` and, per :rfc:`7265`, must be lowercase.

        Parameters:
            name: The jCal name to validate.
            kind: Names the token in the error message, for example,
                ``"property name"``.
            parser: The parser or component to which the name belongs.
            path: The jCal path to ``name``, used to locate it in the error.

        Raises:
            ~error.JCalParsingError: If ``name`` is not a valid lowercase
                iCalendar token.

        See also:
            :meth:`~icalendar.parser.string.validate_token`
        """
        from icalendar.parser.string import validate_token

        try:
            validate_token(name)
        except ValueError:
            raise cls(
                rf"The {kind} must be a valid iCalendar token, matching the "
                rf"regular expression pattern `[\w.-]+`.",
                parser,
                path,
                value=name,
            ) from None
        if name != name.lower():
            raise cls(f"The {kind} must be lowercase.", parser, path, value=name)


__all__ = [
    "BrokenCalendarProperty",
    "ComponentEndMissing",
    "ComponentStartMissing",
    "FeatureWillBeRemovedInFutureVersion",
    "ICalParsingError",
    "IncompleteAlarmInformation",
    "IncompleteComponent",
    "InvalidCalendar",
    "JCalParsingError",
    "LocalTimezoneMissing",
]



"""PERIOD property type from :rfc:`5545`."""

from datetime import date, datetime, timedelta, tzinfo
from typing import Any, ClassVar

from icalendar.compatibility import Self
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.timezone import tzp
from icalendar.tools import is_date, is_datetime, is_pytz, normalize_pytz, to_datetime

from .base import TimeBase
from .datetime import vDatetime
from .duration import vDuration


def _to_midnight(dt: date, tz: tzinfo | None) -> datetime:
    """Convert a date to the datetime at midnight.

    Parameters:
        dt: The date to convert.
        tz: The timezone of the result, or ``None`` for a naive result.

    Returns:
        The datetime at midnight.
    """
    midnight = to_datetime(dt)
    if tz is None:
        return midnight
    if is_pytz(tz):
        return tz.localize(midnight)  # type: ignore[attr-defined]
    return midnight.replace(tzinfo=tz)


def _to_period_datetimes(
    start: date | datetime, end_or_duration: date | datetime | timedelta
) -> tuple[datetime, datetime | timedelta]:
    """Convert the dates of a period to datetimes.

    :rfc:`5545#section-3.3.9` builds a period from datetimes only, but calendars in the
    wild use dates. A date becomes midnight so that the period can be used
    and written back out.

    A converted half takes the timezone of the other half. Without this, an
    aware and a naive datetime could not be subtracted to compute the
    duration.

    Parameters:
        start: The start of the period.
        end_or_duration: The end of the period or its duration.

    Returns:
        The start and the end or duration, with any date converted.
    """
    if is_date(start):
        start = _to_midnight(start, getattr(end_or_duration, "tzinfo", None))
    if is_date(end_or_duration):
        end_or_duration = _to_midnight(end_or_duration, getattr(start, "tzinfo", None))
    return start, end_or_duration


class vPeriod(TimeBase):
    """A span of time, written either as a start and an end or as a start and a duration.

    The value is a tuple of two :class:`datetime.datetime` objects, or of a
    datetime and a :class:`datetime.timedelta`. Whichever way it was written,
    :attr:`start`, :attr:`end`, and :attr:`duration` are all available, and
    :attr:`by_duration` says which of the two forms is written back out.

    Conforming with :rfc:`5545#section-3.3.9`, a period is built from datetimes
    only, the start must come before the end, and a duration must be positive.
    A half written as a date does not conform. Such a half is read as midnight,
    in the timezone of the other half where it has one, so that a calendar that
    gets this wrong can still be read and written.

    ``FREEBUSY`` holds periods directly. ``RDATE`` holds them through
    :class:`~icalendar.prop.dt.list.vDDDLists`, which splits a comma separated
    list and hands each value to :class:`~icalendar.prop.dt.types.vDDDTypes`.
    The halves themselves are parsed by
    :class:`~icalendar.prop.dt.datetime.vDatetime` and
    :class:`~icalendar.prop.dt.duration.vDuration`.

    Parameters:
        per: The start of the period, and either its end or its duration.
        params: The parameters of the property.

    Raises:
        TypeError: If the start is not a date or a datetime, or if the end is
            not a date, a datetime, or a duration.
        ValueError: If the start is after the end.

    Examples:
        A period from 18:00:00 UTC on January 1, 1997 to 07:00:00 UTC on
        January 2, 1997, and one that starts at 18:00:00 UTC and lasts 5 hours
        and 30 minutes:

        .. code-block:: ics

            19970101T180000Z/19970102T070000Z
            19970101T180000Z/PT5H30M

        .. code-block:: pycon

            >>> from icalendar.prop import vPeriod
            >>> period = vPeriod.from_ical("19970101T180000Z/19970102T070000Z")
            >>> vPeriod(period).to_ical()
            b'19970101T180000Z/19970102T070000Z'
            >>> period = vPeriod.from_ical("19970101T180000Z/PT5H30M")
            >>> vPeriod(period).duration
            datetime.timedelta(seconds=19800)

        A half written as a date is read as midnight:

        .. code-block:: pycon

            >>> vPeriod(vPeriod.from_ical("19970101/19970102")).to_ical()
            b'19970101T000000/19970102T000000'

    ..  versionchanged:: 7.2.3

        A period written with dates is read as midnight.
    """

    default_value: ClassVar[str] = "PERIOD"
    params: Parameters
    #: Whether the value is written as a duration rather than as an end.
    by_duration: bool
    #: The start of the period.
    start: datetime
    #: The end of the period, computed from the duration where there is one.
    end: datetime
    #: The time between the start and the end.
    duration: timedelta

    def __init__(
        self,
        per: tuple[date | datetime, date | datetime | timedelta],
        params: dict[str, Any] | None = None,
    ) -> None:
        start, end_or_duration = per
        if not (isinstance(start, (datetime, date))):
            raise TypeError("Start value MUST be a datetime or date instance")
        if not (isinstance(end_or_duration, (datetime, date, timedelta))):
            raise TypeError(
                "end_or_duration MUST be a datetime, date or timedelta instance"
            )
        start, end_or_duration = _to_period_datetimes(start, end_or_duration)
        by_duration = isinstance(end_or_duration, timedelta)
        if by_duration:
            duration = end_or_duration
            end = normalize_pytz(start + duration)
        else:
            end = end_or_duration
            duration = normalize_pytz(end - start)
        if start > end:
            raise ValueError("Start time is greater than end time")

        self.params = Parameters(params or {"value": "PERIOD"})
        # set the timezone identifier
        # does not support different timezones for start and end
        self.params.update_tzid_from(start)

        self.start = start
        self.end = end
        self.by_duration = by_duration
        self.duration = duration

    def overlaps(self, other):
        if self.start > other.start:
            return other.overlaps(self)
        return self.start <= other.start < self.end

    def to_ical(self):
        if self.by_duration:
            return (
                vDatetime(self.start).to_ical()
                + b"/"
                + vDuration(self.duration).to_ical()
            )
        return vDatetime(self.start).to_ical() + b"/" + vDatetime(self.end).to_ical()

    @staticmethod
    def from_ical(ical, timezone=None):
        from icalendar.prop.dt.types import vDDDTypes

        try:
            start, end_or_duration = ical.split("/")
            start = vDDDTypes.from_ical(start, timezone=timezone)
            end_or_duration = vDDDTypes.from_ical(end_or_duration, timezone=timezone)
        except Exception as e:
            raise ValueError(f"Expected period format, got: {ical}") from e
        return _to_period_datetimes(start, end_or_duration)

    def __repr__(self):
        p = (self.start, self.duration) if self.by_duration else (self.start, self.end)
        return f"vPeriod({p!r})"

    @property
    def dt(self):
        """Make this cooperate with the other vDDDTypes."""
        return (self.start, (self.duration if self.by_duration else self.end))

    @property
    def ical_value(self) -> tuple[datetime, timedelta | datetime]:
        """
        Returns the period as a tuple of its start datetime
        and either its end datetime or duration.
        """
        return self.dt

    from icalendar.param import FBTYPE

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vPeriod."""
        return [
            vPeriod((datetime(2025, 11, 10, 16, 35), timedelta(hours=1, minutes=30))),
            vPeriod((datetime(2025, 11, 10, 16, 35), datetime(2025, 11, 10, 18, 5))),
        ]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        value = [vDatetime(self.start).to_jcal(name)[-1]]
        if self.by_duration:
            value.append(vDuration(self.duration).to_jcal(name)[-1])
        else:
            value.append(vDatetime(self.end).to_jcal(name)[-1])
        return [name, self.params.to_jcal(exclude_utc=True), self.VALUE.lower(), value]

    @classmethod
    def parse_jcal_value(
        cls, jcal: str | list
    ) -> tuple[datetime, datetime] | tuple[datetime, timedelta]:
        """Parse a jCal value.

        Raises:
            ~error.JCalParsingError: If the period is not a list with exactly two items,
                or it can't parse a date-time or duration.
        """
        if isinstance(jcal, str) and "/" in jcal:
            # only occurs in the example of RFC7265, Section B.2.2.
            jcal = jcal.split("/")
        if not isinstance(jcal, list) or len(jcal) != 2:
            raise JCalParsingError(
                "A period must be a list with exactly 2 items.", cls, value=jcal
            )
        with JCalParsingError.reraise_with_path_added(0):
            start = vDatetime.parse_jcal_value(jcal[0])
        with JCalParsingError.reraise_with_path_added(1):
            JCalParsingError.validate_value_type(jcal[1], str, cls)
            if jcal[1].startswith(("P", "-P", "+P")):
                end_or_duration = vDuration.parse_jcal_value(jcal[1])
            else:
                try:
                    end_or_duration = vDatetime.parse_jcal_value(jcal[1])
                except JCalParsingError as e:
                    raise JCalParsingError(
                        "Cannot parse date-time or duration.",
                        cls,
                        value=jcal[1],
                    ) from e
        return start, end_or_duration

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265`.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        with JCalParsingError.reraise_with_path_added(3):
            start, end_or_duration = cls.parse_jcal_value(jcal_property[3])
        params = Parameters.from_jcal_property(jcal_property)
        tzid = params.tzid

        if tzid:
            start = tzp.localize(start, tzid)
            if is_datetime(end_or_duration):
                end_or_duration = tzp.localize(end_or_duration, tzid)

        return cls((start, end_or_duration), params=params)


__all__ = ["vPeriod"]



"""BINARY values from :rfc:`5545`."""

import base64
import binascii
from typing import ClassVar

from icalendar.compatibility import Self, deprecate_for_version_8
from icalendar.error import JCalParsingError
from icalendar.parser import Parameters
from icalendar.parser_tools import to_unicode


class vBinary:
    """Binary property values are Base64 encoded."""

    default_value: ClassVar[str] = "BINARY"
    params: Parameters
    bytes: bytes
    """The raw binary value of the BINARY property.

    This is the authoritative storage and round-trips losslessly, including
    for non-UTF-8 data. Use this instead of the deprecated :attr:`obj`.
    """

    def __init__(self, obj: str | bytes, params: dict[str, str] | None = None) -> None:
        if isinstance(obj, str):
            self.bytes = obj.encode("utf-8")
        else:
            self.bytes = obj
        self.params = Parameters(encoding="BASE64", value="BINARY")
        if params:
            self.params.update(params)

    def __repr__(self) -> str:
        return f"vBinary({self.to_ical()})"

    def to_ical(self) -> bytes:
        return base64.b64encode(self.bytes)

    @staticmethod
    def from_ical(ical: str | bytes) -> bytes:
        try:
            return base64.b64decode(ical, validate=True)
        except (binascii.Error, ValueError) as e:
            raise ValueError("Not valid base 64 encoding.") from e

    @property
    def base64data(self) -> str:
        """The Base64-encoded string view of this value.

        This is the same string that :meth:`to_ical` produces, exposed as
        a plain :class:`str` instead of :class:`bytes` so you don't need to
        call :func:`base64.b64encode` yourself. See :issue:`1550`.

        Returns:
            The Base64-encoded representation of the stored value.
        """
        return self.to_ical().decode("ascii")

    @base64data.setter
    def base64data(self, value: str) -> None:
        """Set this value from a Base64-encoded string.

        The decoded raw bytes are stored in :attr:`bytes`, so non-UTF-8
        payloads round-trip losslessly.

        Parameters:
            value: A Base64-encoded string.

        Raises:
            ValueError: If ``value`` isn't valid Base64.
        """
        self.bytes = self.from_ical(value)

    @property
    @deprecate_for_version_8
    def obj(self) -> str:
        """Deprecated string view of the value.

        .. deprecated:: 7.1.3
            Use :attr:`bytes` for the raw binary value. ``obj`` decodes
            :attr:`bytes` as text and is lossy for non-UTF-8 data. It will
            be removed in icalendar 8.
        """
        return to_unicode(self.bytes)

    @obj.setter
    @deprecate_for_version_8
    def obj(self, value: str | bytes) -> None:
        self.bytes = value.encode("utf-8") if isinstance(value, str) else value

    def __eq__(self, other: object) -> bool:
        """self == other"""
        return isinstance(other, vBinary) and self.bytes == other.bytes

    def __hash__(self) -> int:
        """Hash of the vBinary object."""
        return hash(self.bytes)

    @classmethod
    def examples(cls) -> list[Self]:
        """Examples of vBinary."""
        return [cls("VGhlIHF1aWNrIGJyb3duIGZveCBqdW1wcyBvdmVyIHRoZSBsYXp5IGRvZy4")]

    from icalendar.param import VALUE

    def to_jcal(self, name: str) -> list:
        """The jCal representation of this property according to :rfc:`7265`."""
        params = self.params.to_jcal()
        if params.get("encoding") == "BASE64":
            # BASE64 is the only allowed encoding
            del params["encoding"]
        return [name, params, self.VALUE.lower(), self.base64data]

    @property
    def ical_value(self) -> bytes:
        """The raw ``bytes`` value of the BINARY property.

        .. versionadded:: 7.1.0

        .. versionchanged:: 7.1.3
            Returns the raw stored bytes. Previously the stored value was
            Base64-decoded, which raised :class:`ValueError` for non-Base64
            input. See :pr:`1356`.
        """
        return self.bytes

    @classmethod
    def from_jcal(cls, jcal_property: list) -> Self:
        """Parse jCal from :rfc:`7265` to a vBinary.

        Parameters:
            jcal_property: The jCal property to parse.

        Raises:
            ~error.JCalParsingError: If the provided jCal is invalid.
        """
        JCalParsingError.validate_property(jcal_property, cls)
        JCalParsingError.validate_value_type(jcal_property[3], str, cls, 3)
        return cls(
            cls.from_ical(jcal_property[3]),
            params=Parameters.from_jcal_property(jcal_property),
        )


__all__ = ["vBinary"]



""":rfc:`5545` iCalendar component."""

from __future__ import annotations

import uuid
from datetime import timedelta
from typing import TYPE_CHECKING, Literal, cast, overload

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    categories_property,
    images_property,
    multi_language_text_property,
    single_string_property,
    source_property,
    uid_property,
    url_property,
)
from icalendar.cal.component import Component
from icalendar.cal.examples import get_example
from icalendar.cal.timezone import Timezone
from icalendar.error import IncompleteComponent
from icalendar.parser.ical.calendar import CalendarIcalParser
from icalendar.version import __version__

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from datetime import date, datetime
    from pathlib import Path

    from icalendar.cal.availability import Availability
    from icalendar.cal.event import Event
    from icalendar.cal.free_busy import FreeBusy
    from icalendar.cal.journal import Journal
    from icalendar.cal.todo import Todo
    from icalendar.compatibility import Self
    from icalendar.parser.ical.component import ComponentIcalParser


DEFAULT_PRODID = f"-//collective//icalendar//{__version__}//EN"


class Calendar(Component):
    """
        The "VCALENDAR" object is a collection of calendar information.
        This information can include a variety of components, such as
        "VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VTIMEZONE", or any
        other type of calendar component.

    Examples:
        Create a new Calendar:

            >>> from icalendar import Calendar
            >>> calendar = Calendar.new(name="My Calendar")
            >>> print(calendar.calendar_name)
            My Calendar

    """

    name = "VCALENDAR"
    canonical_order = (
        "VERSION",
        "PRODID",
        "CALSCALE",
        "METHOD",
        "DESCRIPTION",
        "X-WR-CALDESC",
        "NAME",
        "X-WR-CALNAME",
    )
    required = (
        "PRODID",
        "VERSION",
    )
    singletons = (
        "PRODID",
        "VERSION",
        "CALSCALE",
        "METHOD",
        "COLOR",  # RFC 7986
    )
    multiple = (
        "CATEGORIES",  # RFC 7986
        "DESCRIPTION",  # RFC 7986
        "NAME",  # RFC 7986
    )

    @classmethod
    def example(cls, name: str = "example") -> Calendar:
        """Return the calendar example with the given name."""
        return cls.from_ical(get_example("calendars", name))

    @classmethod
    def _get_ical_parser(cls, st: str | bytes) -> ComponentIcalParser:
        """Get the iCal parser for the given input string."""
        return CalendarIcalParser(st, cls._get_component_factory(), cls.types_factory)

    @overload
    @classmethod
    def from_ical(
        cls, st: str | bytes | Path, multiple: Literal[False] = False
    ) -> Calendar: ...

    @overload
    @classmethod
    def from_ical(
        cls, st: str | bytes | Path, multiple: Literal[True]
    ) -> list[Calendar]: ...

    @classmethod
    def from_ical(
        cls, st: str | bytes | Path, multiple: bool = False
    ) -> Calendar | list[Calendar]:
        """Parse iCalendar data into calendar instances.

        Parameters:
            st: iCalendar data as bytes or string, or a path to an iCalendar file.
            multiple: If ``True``, returns a list of calendars.
                If ``False``, returns a single calendar.

        Returns:
            Calendar or list of calendars.
        """
        return cast(
            "Calendar | list[Calendar]", super().from_ical(st, multiple=multiple)
        )

    @property
    def events(self) -> list[Event]:
        """All event components in the calendar.

        This is a shortcut to get all events.
        Modifications do not change the calendar.
        Use :meth:`Component.add_component <icalendar.cal.component.Component.add_component>`.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example()
        >>> event = calendar.events[0]
        >>> event.start
        datetime.date(2022, 1, 1)
        >>> print(event["SUMMARY"])
        New Year's Day
        """
        return self.walk("VEVENT")

    @property
    def todos(self) -> list[Todo]:
        """All todo components in the calendar.

        This is a shortcut to get all todos.
        Modifications do not change the calendar.
        Use :meth:`Component.add_component <icalendar.cal.component.Component.add_component>`.
        """
        return self.walk("VTODO")

    @property
    def journals(self) -> list[Journal]:
        """All journal components in the calendar.

        This is a shortcut to get all journals.
        Modifications do not change the calendar.
        Use :meth:`Component.add_component <icalendar.cal.component.Component.add_component>`.
        """
        return self.walk("VJOURNAL")

    @property
    def availabilities(self) -> list[Availability]:
        """All :class:`Availability` components in the calendar.

        This is a shortcut to get all availabilities.
        Modifications do not change the calendar.
        Use :meth:`Component.add_component <icalendar.cal.component.Component.add_component>`.
        """
        return self.walk("VAVAILABILITY")

    @property
    def freebusy(self) -> list[FreeBusy]:
        """All FreeBusy components in the calendar.

        This is a shortcut to get all FreeBusy.
        Modifications do not change the calendar.
        Use :meth:`Component.add_component <icalendar.cal.component.Component.add_component>`.
        """
        return self.walk("VFREEBUSY")

    def get_used_tzids(self) -> set[str]:
        """The set of TZIDs in use.

        This goes through the whole calendar to find all occurrences of
        timezone information like the TZID parameter in all attributes.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example("timezone_rdate")
        >>> calendar.get_used_tzids()
        {'posix/Europe/Vaduz'}

        Even if you use UTC, this will not show up.
        """
        result = set()
        for _name, value in self.property_items(sorted=False):
            if hasattr(value, "params"):
                result.add(value.params.get("TZID"))
        return result - {None}

    def get_missing_tzids(self) -> set[str]:
        """The set of missing timezone component tzids.

        To create a :rfc:`5545` compatible calendar,
        all of these timezones should be added.

        UTC is excluded: per :rfc:`5545#section-3.2.19`, UTC datetimes use
        the ``Z`` suffix and never require a VTIMEZONE component.
        """
        tzids = self.get_used_tzids() - {"UTC"}
        for timezone in self.timezones:
            # discard (not remove) — a VTIMEZONE may exist for a timezone not
            # referenced by any event TZID (e.g. added by x-wr-timezone conversion)
            tzids.discard(timezone.tz_name)
        return tzids

    @property
    def timezones(self) -> list[Timezone]:
        """Return the timezones components in this calendar.

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example("pacific_fiji")
        >>> [timezone.tz_name for timezone in calendar.timezones]
        ['custom_Pacific/Fiji']

        .. note::

            This is a read-only property.
        """
        return self.walk("VTIMEZONE")

    def add_missing_timezones(
        self,
        first_date: date = Timezone.DEFAULT_FIRST_DATE,
        last_date: date = Timezone.DEFAULT_LAST_DATE,
    ):
        """Add all missing VTIMEZONE components.

        This adds all the timezone components that are required.
        VTIMEZONE components are inserted at the beginning of the calendar
        to ensure they appear before other components that reference them.

        .. note::

            Timezones that are not known will not be added.

        Parameters:
            first_date: Earlier than anything that happens in the calendar.
            last_date: Later than anything happening in the calendar.

        >>> from icalendar import Calendar, Event
        >>> from datetime import datetime
        >>> from zoneinfo import ZoneInfo
        >>> calendar = Calendar()
        >>> event = Event()
        >>> calendar.add_component(event)
        >>> event.start = datetime(1990, 10, 11, 12, tzinfo=ZoneInfo("Europe/Berlin"))
        >>> calendar.timezones
        []
        >>> calendar.add_missing_timezones()
        >>> calendar.timezones[0].tz_name
        'Europe/Berlin'
        >>> calendar.get_missing_tzids()  # check that all are added
        set()
        """
        missing_tzids = self.get_missing_tzids()
        if not missing_tzids:
            return

        existing_timezone_count = len(self.timezones)

        for tzid in missing_tzids:
            try:
                timezone = Timezone.from_tzid(
                    tzid, first_date=first_date, last_date=last_date
                )
            except ValueError:
                continue
            self.subcomponents.insert(existing_timezone_count, timezone)
            existing_timezone_count += 1

    calendar_name = multi_language_text_property(
        "NAME",
        "X-WR-CALNAME",
        """The display name of this calendar, per :rfc:`7986#section-5.1`.

    Implements both the ``NAME`` property from :rfc:`7986#section-5.1` and the widely used
    ``X-WR-CALNAME`` extension for broader client compatibility.

    Multiple language variants can be stored by setting this property more than
    once, each with a different ``LANGUAGE`` parameter value.

    Example:
        Set the name of the calendar.

        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.calendar_name = "My Calendar"
            >>> print(calendar.to_ical().decode())
            BEGIN:VCALENDAR
            NAME:My Calendar
            X-WR-CALNAME:My Calendar
            END:VCALENDAR

    """,
    )

    description = multi_language_text_property(
        "DESCRIPTION",
        "X-WR-CALDESC",
        """A description of the calendar's content.

    Implements both ``DESCRIPTION`` from :rfc:`5545#section-3.8.1.5` and
    :rfc:`7986#section-5.2` and ``X-WR-CALDESC`` for broader calendar client
    compatibility.

    Multiple language variants can be stored by setting this property more than
    once with different ``LANGUAGE`` parameter values.

    Example:
        Add a description to a calendar.

        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.description = "This is a calendar"
            >>> print(calendar.to_ical().decode())
            BEGIN:VCALENDAR
            DESCRIPTION:This is a calendar
            X-WR-CALDESC:This is a calendar
            END:VCALENDAR

    """,
    )

    color = single_string_property(
        "COLOR",
        """A CSS3 color name or value used to visually distinguish this calendar, per :rfc:`7986#section-5.9`.

    Implements both ``COLOR`` from :rfc:`7986#section-5.9` and ``X-APPLE-CALENDAR-COLOR``.
    The value is a case-insensitive CSS3 color name, for example, ``"turquoise"``, or
    a hex code, for example, ``"#ffffff"``, drawn from the
    `CSS3 color specification <https://www.w3.org/TR/css-color-3/>`_.

    Since :rfc:`7986`, individual ``VEVENT``, ``VTODO``, and ``VJOURNAL``
    subcomponents may also carry their own color.

    Example:
        .. code-block:: pycon

            >>> from icalendar import Calendar
            >>> calendar = Calendar()
            >>> calendar.color = "black"
            >>> print(calendar.to_ical().decode())
            BEGIN:VCALENDAR
            COLOR:black
            END:VCALENDAR

    """,
        "X-APPLE-CALENDAR-COLOR",
    )
    categories = categories_property
    uid = uid_property
    prodid = single_string_property(
        "PRODID",
        """The product identifier for the software that created this iCalendar object.

This property is defined in :rfc:`5545#section-3.7.3`.
It's required exactly once per iCalendar object.

The value should be a globally unique string. The conventional format is a
Formal Public Identifier (FPI), for example, ``-//My Company//My Product//EN``, but any
unique string is acceptable.

Example:
    Set a custom product identifier on a new calendar.

    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> cal = Calendar()
        >>> cal.prodid = "-//MyApp//MyCalendar//EN"
        >>> str(cal.prodid)
        '-//MyApp//MyCalendar//EN'

See also:
    :attr:`version`
""",
    )
    version = single_string_property(
        "VERSION",
        """The iCalendar specification version required to interpret this object.

This property is defined in :rfc:`5545#section-3.7.4`.
It's required exactly once per calendar object.
The value ``"2.0"`` indicates :rfc:`5545` compliance, which is the default used
by this library. A range such as ``"1.0;2.0"`` may indicate minimum and maximum
supported versions.

Example:
    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> cal = Calendar()
        >>> cal.version = "2.0"
        >>> str(cal.version)
        '2.0'

See also:
    :attr:`prodid`
""",
    )

    calscale = single_string_property(
        "CALSCALE",
        """The calendar scale for date and time values in this iCalendar object.

This property is defined in :rfc:`5545#section-3.7.1`. The only value currently defined is
``"GREGORIAN"`` (the default). When this property is absent, Gregorian is assumed.

Per :rfc:`7529`, non-Gregorian calendar systems are expressed via ``RRULE``
transformations rather than a different ``CALSCALE`` value. However, icalendar
currently implements only the parsing of this value, not a transformation.

Example:
    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> cal = Calendar()
        >>> cal.calscale
        'GREGORIAN'

        """,
        default="GREGORIAN",
    )
    method = single_string_property(
        "METHOD",
        """The iTIP scheduling method associated with this calendar object, per :rfc:`5545#section-3.7.2`.

When present, ``METHOD`` indicates that this object is part of a scheduling
transaction, such as a meeting invitation or cancellation. Scheduling methods
are defined by :rfc:`5546#section-1.4` (iTIP), with values such as ``"REQUEST"``,
``"REPLY"``, ``"CANCEL"``, and ``"PUBLISH"``.

When used inside a MIME message, this value must match the ``method`` parameter
of the ``Content-Type`` header. If absent, the calendar is treated as a plain
data snapshot with no scheduling semantics.

Example:
    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> cal = Calendar()
        >>> cal.method = "REQUEST"
        >>> str(cal.method)
        'REQUEST'

""",
    )
    url = url_property
    source = source_property

    @property
    def refresh_interval(self) -> timedelta | None:
        """A suggested minimum polling interval for fetching updates to this calendar, per :rfc:`7986#section-5.7`.

        Calendar clients should not poll more frequently than this interval.
        The value must be a positive duration.

        Returns:
            A :class:`~datetime.timedelta`, or ``None`` when not set.

        Raises:
            ValueError: When setting a non-positive (zero or negative) duration.
            TypeError: When setting a value that is not a :class:`~datetime.timedelta` or ``None``.

        Example:
            .. code-block:: pycon

                >>> from datetime import timedelta
                >>> from icalendar import Calendar
                >>> cal = Calendar()
                >>> cal.refresh_interval = timedelta(hours=1)
                >>> cal.refresh_interval
                datetime.timedelta(seconds=3600)

        """
        refresh_interval = self.get("REFRESH-INTERVAL")
        return refresh_interval.dt if refresh_interval else None

    @refresh_interval.setter
    def refresh_interval(self, value: timedelta | None):
        """Set the REFRESH-INTERVAL."""
        if not isinstance(value, timedelta) and value is not None:
            raise TypeError(
                "REFRESH-INTERVAL must be either a positive timedelta,"
                " or None to delete it."
            )
        if value is not None and value.total_seconds() <= 0:
            raise ValueError("REFRESH-INTERVAL must be a positive timedelta.")
        if value is not None:
            del self.refresh_interval
            self.add("REFRESH-INTERVAL", value)
        else:
            del self.refresh_interval

    @refresh_interval.deleter
    def refresh_interval(self):
        """Delete REFRESH-INTERVAL."""
        self.pop("REFRESH-INTERVAL")

    images = images_property

    @classmethod
    def new(
        cls,
        /,
        calscale: str | None = None,
        categories: Sequence[str] = (),
        color: str | None = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        description: str | None = None,
        language: str | None = None,
        last_modified: date | datetime | None = None,
        links: LINKS_TYPE_SETTER = None,
        method: str | None = None,
        name: str | None = None,
        organization: str | None = None,
        prodid: str | None = None,
        refresh_interval: timedelta | None = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        source: str | None = None,
        subcomponents: Iterable[Component] | None = None,
        uid: str | uuid.UUID | None = None,
        url: str | None = None,
        version: str = "2.0",
    ) -> Self:
        """Create a new Calendar.

        This creates a new Calendar in accordance with :rfc:`5545` and :rfc:`7986`.

        Parameters:
            calscale: The :attr:`calscale` of the calendar.
            categories: The :attr:`categories` of the calendar.
            color: The :attr:`color` of the calendar.
            concepts: The :attr:`~icalendar.cal.component.Component.concepts` of the calendar.
            description: The :attr:`description` of the calendar.
            language: The language for the calendar. Used to generate localized `prodid`.
            last_modified: The :attr:`~icalendar.cal.component.Component.last_modified` of the calendar.
            links: The :attr:`~icalendar.cal.component.Component.links` of the calendar.
            method: The :attr:`method` of the calendar.
            name: The :attr:`calendar_name` of the calendar.
            organization: The organization name. Used to generate `prodid` if not provided.
            prodid: The :attr:`prodid` of the component. If ``None`` and ``organization`` is provided,
                generates a `prodid` in the format of "-//organization//name//language".
                If ``None`` and ``organization`` is not provided, sets it to
                :attr:`~icalendar.cal.calendar.DEFAULT_PRODID`.
            refresh_interval: The :attr:`refresh_interval` of the calendar.
            refids: :attr:`~icalendar.cal.component.Component.refids` of the calendar.
            related_to: :attr:`~icalendar.cal.component.Component.related_to` of the calendar.
            source: The :attr:`source` of the calendar.
            subcomponents: The subcomponents of the calendar.
            uid: The :attr:`uid` of the calendar.
                If None, this is set to a new :func:`uuid.uuid4`.
            url: The :attr:`url` of the calendar.
            version: The :attr:`version` of the calendar.

        Returns:
            :class:`Calendar`

        Raises:
            ~error.InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        calendar: Self = super().new(
            last_modified=last_modified,
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
            subcomponents=subcomponents,
        )

        # Generate prodid if not provided but organization is given
        if prodid is None and organization:
            app_name = name or "Calendar"
            lang = language.upper() if language else "EN"
            prodid = f"-//{organization}//{app_name}//{lang}"
        elif prodid is None:
            prodid = DEFAULT_PRODID

        calendar.prodid = prodid
        calendar.version = version
        calendar.calendar_name = name
        calendar.color = color
        calendar.description = description
        calendar.method = method
        calendar.calscale = calscale
        calendar.categories = categories
        calendar.uid = uid if uid is not None else uuid.uuid4()
        calendar.url = url
        calendar.refresh_interval = refresh_interval
        calendar.source = source

        return calendar

    def validate(self):
        """Validate that the calendar has required properties and components.

        This method can be called explicitly to validate a calendar before output.

        Raises:
            ~error.IncompleteComponent: If the calendar lacks required properties or
                components.
        """
        if not self.get("PRODID"):
            raise IncompleteComponent("Calendar must have a PRODID")
        if not self.get("VERSION"):
            raise IncompleteComponent("Calendar must have a VERSION")
        if not self.subcomponents:
            raise IncompleteComponent(
                "Calendar must contain at least one component (event, todo, etc.)"
            )


__all__ = ["Calendar"]
